import os
import sys
from typing import Optional

import openai
import typer
from rich.console import Console
from youtube_transcript_api import TranscriptsDisabled

# from .config import load_config, Config
from chat_gpt.commands.chat import chat
from chat_gpt.commands.youtube import chat_with_yt_video

app = typer.Typer()

console = Console()

openai.api_key = os.environ.get("OPENAI_API_KEY")

if openai.api_key is None:
    console.print("[red]Please set the OPENAI_API_KEY environment variable.")
    exit(1)


@app.command("chat")
def start(
    model_name: str = "gpt-4o", markdown: bool = False, file_path: Optional[str] = None
):
    """
    Start conversation with our assistant
    """
    try:
        chat(model_name, markdown=markdown, file_path=file_path)
    except KeyboardInterrupt:
        try:
            sys.exit(130)
        except SystemExit:
            os._exit(130)


@app.command()
def youtube(url: str, language: str = "en"):
    """
    Start conversation with our assistant using a youtube video transcript
    """

    try:
        chat_with_yt_video(url, language)
    except TranscriptsDisabled as e:
        console.print(
            "\n[red]Transcripts are disabled for this video or the video you're trying to extract doesn't exist."
        )
        console.print(e)
    except KeyboardInterrupt:
        try:
            sys.exit(130)
        except SystemExit:
            os._exit(130)


#
#
# @app.command()
# def show_config():
#     config: Config = load_config()
#
#     for key, value in config.model_dump().items():
#         typer.echo(f"{key}: {value}")
#
#
# @app.command()
# def rag():
#     import questionary
#
#     p = questionary.path(
#         "Enter the path to the RAG model file:", only_directories=True
#     ).ask()
#     print(p)
#     print("type", type(p))
#
#     action = questionary.rawselect(
#         "Choose an action:",
#         choices=["Chat", "RAG", "Image", "STOP"],
#     ).ask()
#
#     match action:
#         case "Chat":
#             start()
#         case "RAG":
#             typer.echo("RAG")
#         case "Image":
#             typer.echo("Image")
#             image_action = questionary.rawselect(
#                 "Choose an action:", choices=["Last screenshot", "From file path"]
#             )
#
#             match image_action:
#                 case "Last screenshot":
#                     typer.echo("Last screenshot")
#                 case "From file path":
#                     typer.echo("From file path")
#                     file_path = questionary.path(
#                         "Enter the path to the image file:", only_files=True
#                     ).ask()
#         case "STOP":
#             typer.echo("STOP")
#             sys.exit(0)



from typing import Any, Literal

from langchain.schema import Document
from langchain_weaviate import WeaviateVectorStore


class WeaviateVectorStoreWrapper(WeaviateVectorStore):
    """Wrapper around WeaviateVectorStore to handle Parent/Child indexes.
    We perform the search on the child index and return the parent index for the LLM.
    """

    import weaviate
    import weaviate.classes as wvc

    def __init__(
        self,
        client: weaviate.WeaviateClient,
        index_name: str,
        text_key: str,
        attributes: list[str] | None = None,
        references: list[wvc.query.QueryReference] | None = None,
        ref_attributes: list[str] | None = None,
    ):
        """Initialize with Weaviate client and search_kwargs."""
        super().__init__(
            client=client,
            index_name=index_name,
            text_key=text_key,
            attributes=attributes,
        )

        self.references = references
        self.parent_attributes = ref_attributes

    def _perform_search(
        self,
        query: str,
        k: int,
        return_score=False,
        search_method: Literal[
            "similarity_search", "hybrid", "near_vector"
        ] = "similarity_search",
        tenant: str | None = None,
        **kwargs: Any,
    ) -> list[Document]:
        """
        Perform a similarity search.

        Parameters:
        query (str): The query string to search for.
        k (int): The number of results to return.
        return_score (bool, optional): Whether to return the score along with the
          document. Defaults to False.
        search_method (Literal['hybrid', 'near_vector'], optional): The search method
          to use. Can be 'hybrid' or 'near_vector'. Defaults to 'hybrid'.
        tenant (Optional[str], optional): The tenant name. Defaults to None.
        **kwargs: Additional parameters to pass to the search method. These parameters
          will be directly passed to the underlying Weaviate client's search method.

        Returns:
        List[Union[Document, Tuple[Document, float]]]: A list of documents that match
          the query. If return_score is True, each document is returned as a tuple
          with the document and its score.

        Raises:
        ValueError: If _embedding is None or an invalid search method is provided.
        """
        import weaviate

        with self._tenant_context(tenant) as collection:
            try:
                if search_method == "similarity_search":
                    result = collection.query.near_text(query=query, limit=k, **kwargs)
                elif search_method == "hybrid":
                    result = collection.query.hybrid(query=query, limit=k, **kwargs)
                elif search_method == "near_vector":
                    result = collection.query.near_vector(limit=k, **kwargs)
                else:
                    raise ValueError(f"Invalid search method: {search_method}")
            except weaviate.exceptions.WeaviateQueryException as e:
                raise ValueError(f"Error during query: {e}")

        assert len(result.objects) > 0, "The search did not return any results."

        docs = []
        for obj in result.objects:
            text = obj.properties.pop(self._text_key)
            filtered_metadata = {
                k: v
                for k, v in obj.metadata.__dict__.items()
                if v is not None and k != "score"
            }
            merged_props = {
                "id": str(obj.uuid),
                **obj.properties,
                **filtered_metadata,
                **({"vector": obj.vector["default"]} if obj.vector else {}),
            }

            if obj.references:
                for ref in obj.references:
                    merged_props[ref] = obj.references[ref].objects[
                        0
                    ]  # HACK: we suppose only links to one object for a given ref

            doc = Document(page_content=text, metadata=merged_props)
            if not return_score:
                docs.append(doc)
            else:
                score = obj.metadata.score
                docs.append((doc, score))

        return docs

    def similarity_search(
        self, query: str, k: int = 4, **kwargs: Any
    ) -> list[Document]:
        """Return docs most similar to query.

        Args:
            query: Text to look up documents similar to.
            k: Number of Documents to return. Defaults to 4.

        Returns:
            List of Documents most similar to the query.
        """

        if self.references is None:
            ########################################################################################
            # Perform similarity search on a single index
            ########################################################################################

            return self._perform_search(query, k, **kwargs)

        else:
            ########################################################################################
            # Perform similarity search on a child index, but return parent index for the LLM
            ########################################################################################

            if not self.parent_attributes:
                raise ValueError(
                    "You must provide parent attributes when using references"
                )

            if len(self.references) > 1:
                raise ValueError(
                    "WeaviateWithFilter can only perform similarity search with one reference at the moment"
                )

            # used by the similarity search method to return the references (here the parent document of the child document)
            kwargs |= {"return_references": self.references}

            # We perform the search on the child index. We add 20 to the k to make sure we get enough documents
            child_docs = self._perform_search(query, k + 20, **kwargs)

            ########################################################################################
            # We retrieve the parent documents from the child documents (and format it for langchain)
            ########################################################################################

            # We retrieve the parent documents from the child documents
            # We only keep the first k documents of the parent documents and avoid duplicates
            # Each retrieved document has a single parent document, hence the [0] index
            parent_index_name = self.references[0].link_on
            unique_pages = set()
            parent_docs = []

            for doc in child_docs:
                ########################################################################################
                # Store unique pages to avoid duplicates
                ########################################################################################

                if doc.metadata["page"] in unique_pages:
                    continue
                unique_pages.add(doc.metadata["page"])

                ########################################################################################
                # Create metadata
                ########################################################################################

                properties = doc.metadata[parent_index_name].properties
                metadata = {}

                for field in self.parent_attributes:
                    if field == "content":
                        continue
                    elif field == "id":
                        metadata[field] = str(doc.metadata[parent_index_name].uuid)
                    else:
                        metadata[field] = properties[field]

                ########################################################################################
                # Create new Parent document
                ########################################################################################

                new_doc = Document(
                    page_content=properties["content"], metadata=metadata
                )

                ########################################################################################
                # Ensures we don't return more than k documents
                ########################################################################################

                parent_docs.append(new_doc)

                if len(parent_docs) == k:
                    break

            return parent_docs



from operator import itemgetter
from langchain_core.prompts import ChatPromptTemplate
from langchain_community.document_loaders import YoutubeLoader
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain_community.embeddings.openai import OpenAIEmbeddings
from langchain_community.vectorstores import FAISS
from langchain_openai import ChatOpenAI
from langchain_core.output_parsers import StrOutputParser
from rich.console import Console
from rich.prompt import Prompt


def create_db_from_youtube_video_url(
    url: str, language: str, console: Console
) -> FAISS:
    embeddings = OpenAIEmbeddings()

    console.print("[cyan]Start by checking if transcripts availabled ...")

    loader = YoutubeLoader.from_youtube_url(url, language=language)
    console.print("[cyan]Import transcripts recordings ...")

    transcript = loader.load()

    console.print("[cyan]Transcripts received  ...")

    console.print("[cyan]Start splitting transcripts ...")

    text_splitter = RecursiveCharacterTextSplitter(chunk_size=1500, chunk_overlap=100)
    docs = text_splitter.split_documents(transcript)

    console.print("[cyan]Documents created !")
    console.print()
    return FAISS.from_documents(docs, embeddings)


def get_response_from_query(db: FAISS, query: str, k: int, console: Console):
    console = Console()

    # docs = db.similarity_search(query, k=k)
    # docs_page_content = " ".join([d.page_content for d in docs])

    llm = ChatOpenAI(temperature=0.1)

    template = """
        You are a helpful assistant that that can answer questions about youtube videos 
        based on the video's transcript.
        
        Answer the following question: {question}
        By searching the following video transcript: {docs}
        
        Only use the factual information from the transcript to answer the question.
        
        If you feel like you don't have enough information to answer the question, say "I don't know".
        
        Your answers should be verbose and detailed.
    """

    prompt = ChatPromptTemplate.from_template(template)

    chain = (
        {
            "docs": itemgetter("question") | db.as_retriever(),
            "question": itemgetter("question"),
        }
        | prompt
        | llm
        | StrOutputParser()
    )

    console.print("[blue][b]Assistant :[/b][/blue]")

    for chunk in chain.stream({"question": query}):
        console.print(chunk, end="", style="green")


def chat_with_yt_video(url: str, language: str):
    console = Console()

    # video_url = "https://www.youtube.com/watch?v=NYSWn1ipbgg" "https://www.youtube.com/watch?v=L_Guz73e6fw"
    db = create_db_from_youtube_video_url(url, language, console)

    console.print("[green]You can now ask questions about the video's transcript.\n")

    while True:
        content = Prompt.ask("[red][b]User[b/]")
        console.print()

        get_response_from_query(db, content, k=5, console=console)

        console.print()
        console.print()



import os
from pathlib import Path
import yaml
from pydantic import BaseModel


class Config(BaseModel):
    MODEL_NAME: str
    SQLITE_DB: str

    WEAVIATE_HOST: str
    WEAVIATE_PORT: int
    WEAVIATE_GRPC_HOST: str
    WEAVIATE_GRPC_PORT: int


def load_config() -> Config:
    config_path = Path(os.path.expanduser("~/.config/utils/custom-chat-gpt/config.yml"))
    with open(config_path, "r") as file:
        config_data = yaml.safe_load(file)
        config_data = {k.upper(): v for k, v in config_data.items()}
    return Config(**config_data)



from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import ChatPromptTemplate
from langchain_openai import ChatOpenAI
from langchain.schema import AIMessage, BaseMessage, HumanMessage, SystemMessage
from rich.console import Console
from rich.prompt import Prompt
from rich.markdown import Markdown
from rich.live import Live


def chat(model_name: str, markdown: bool, file_path: str | None = None):
    console = Console()

    messages: list[BaseMessage] = [
        SystemMessage(
            content="You’re a helpful programming assistant. Answers the questions as a professional programmer."
        ),
    ]

    console.print("Starting a chat ...")
    console.print()

    user_content = None

    if file_path:
        with open(file_path, "r") as f:
            user_content = f.read()
            console.print("Reading from file ...")

    while True:
        if user_content is None:
            user_content = Prompt.ask("[red][b]User [b/]")

        console.print()
        messages.append(HumanMessage(content=user_content))
        prompt = ChatPromptTemplate.from_messages(messages)

        model = ChatOpenAI(model=model_name, temperature=0.1)

        chain = prompt | model | StrOutputParser()

        ai_content = ""

        if markdown:
            console.print("[blue][b]Assistant :[/b][/blue]")
            with Live("", refresh_per_second=10, console=console) as live:
                for chunk in chain.stream({}):
                    ai_content += chunk
                    live.update(Markdown(ai_content))
        else:
            console.print("[blue][b]Assistant :[/b][/blue]", end=" ")

            for chunk in chain.stream({}):
                ai_content += chunk
                console.print(chunk, end="", style="green")

        messages.append(AIMessage(content=ai_content))
        console.print()
        console.print()

        user_content = None



import os
import weaviate
import weaviate.classes
from chat_gpt.config import Config


def get_weaviate_client(settings: Config) -> weaviate.WeaviateClient:
    return weaviate.connect_to_custom(
        http_host=settings.WEAVIATE_HOST,
        http_port=settings.WEAVIATE_PORT,
        http_secure=False,
        grpc_host=settings.WEAVIATE_GRPC_HOST,
        grpc_port=settings.WEAVIATE_GRPC_PORT,
        grpc_secure=False,
        headers={"X-OpenAI-Api-Key": os.environ["OPENAI_API_KEY"]},
    )



from typing import Type
from collections import OrderedDict
from langchain.schema import Document
from langchain_community.document_loaders.pdf import BasePDFLoader
from langchain.text_splitter import RecursiveCharacterTextSplitter
from pydantic import BaseModel
from weaviate.client import WeaviateClient


class DocumentInfo(BaseModel):
    source: str
    filename: str
    metadata: dict


def format_documents(
    documents: list[Document], document_info: DocumentInfo
) -> list[dict]:
    """
    Documents are extracted via Langchain and contains document specific metadata (page, filename ...).
    Whereas DocumentInfo contains metadata that is common to all documents (company, entity_uid ...) that WE specify.
    Python starts indexing at 0, but we want to start at 1 for the page number.
    """

    formatted_documents = []

    for doc in documents:
        doc_metadata = {k: v for k, v in doc.metadata.items() if k != "source"}
        if "page" in doc_metadata:
            doc_metadata["page"] = int(doc_metadata["page"]) + 1

        formatted_documents.append(
            {
                "content": doc.page_content,
                "filename": document_info.filename,
                **document_info.metadata,
                **doc_metadata,
            }
        )
    return formatted_documents


class DataIngestor:
    """
    Class to ingest PDF documents into Weaviate.
    """

    def __init__(
        self,
        weaviate_client: WeaviateClient,
        text_splitter: RecursiveCharacterTextSplitter,
        document_loader: Type[BasePDFLoader],
    ):
        self.client: WeaviateClient = weaviate_client
        self.text_splitter: RecursiveCharacterTextSplitter = text_splitter
        self.document_loader: Type[BasePDFLoader] = document_loader

    def ingest_documents_with_references(
        self,
        parent_class_name: str,
        child_class_name: str,
        refererence_name: str,
        document_info: DocumentInfo,
    ):
        """Ingests a list of documents into Weaviate and creates a reference between the parent and child class."""

        """
        Using this method in a multi threading environment will likely generate warnings and errors in Weaviate due to gRPC
        which are likely not to be a problem but are annoying.
        """

        from weaviate.util import generate_uuid5

        print("Starting to ingest documents.")

        parent_collection = self.client.collections.get(parent_class_name)

        child_docs_len = 0
        child_docs_with_parent_uuid: OrderedDict[str, list[dict]] = OrderedDict()

        with parent_collection.batch.dynamic() as batch:
            parent_docs: list[Document] = self.document_loader(
                document_info.source
            ).load()
            parent_docs_formatted: list[dict] = format_documents(
                parent_docs, document_info
            )

            print(f"Ingesting {document_info.filename}")

            print(
                f"A total of {len(parent_docs_formatted)} documents to ingest for the Parent class."
            )

            for idx, doc in enumerate(parent_docs_formatted, start=1):
                if idx % 50 == 0 or idx == len(parent_docs_formatted):
                    print(f"Ingesting {idx} of {len(parent_docs_formatted)}")

                parent_uuid: str = generate_uuid5(doc)
                batch.add_object(doc, uuid=parent_uuid)

                new_doc: dict = {k: v for k, v in doc.items() if k != "content"}
                new_doc[
                    "page"
                ] -= 1  # Substract 1 since the parent already has the page number incremented by 1 and the format will add 1 again.

                child_docs: list[Document] = self.text_splitter.split_documents(
                    [Document(page_content=doc["content"], metadata=new_doc)]
                )
                child_docs_formatted: list[dict] = format_documents(
                    child_docs, document_info
                )

                child_docs_len += len(child_docs_formatted)

                child_docs_with_parent_uuid[parent_uuid] = child_docs_formatted

        parent_failed_objects = self.client.batch.failed_objects

        print(f"parent failed_objects: {parent_failed_objects}")

        if parent_failed_objects:
            print(
                f"Failed to ingest {len(parent_failed_objects)} objects in Parent class.",
            )
            for obj in parent_failed_objects:
                print(obj)

        print(f"Creating {child_docs_len} documents for the Child class.")

        print(f"len(child_docs_with_parent_uuid): {len(child_docs_with_parent_uuid)}")

        child_collection = self.client.collections.get(child_class_name)

        with child_collection.batch.dynamic() as batch:
            for parent_uuid, docs in child_docs_with_parent_uuid.items():
                for doc in docs:
                    batch.add_object(
                        properties=doc, references={refererence_name: parent_uuid}
                    )

        child_failed_objects = self.client.batch.failed_objects
        child_failed_references = self.client.batch.failed_references

        if child_failed_objects:
            print(f"Failed to ingest {len(child_failed_objects)} objects.")
            for obj in child_failed_objects:
                print(obj)
                print(obj.object_.collection)
                print(obj.object_.properties)

        if child_failed_references:
            print(
                f"Failed to create {len(child_failed_references)} references.",
            )
            for ref in child_failed_references:
                print(ref)

        if not child_failed_objects and not child_failed_references:
            print("All documents and references were successfully ingested.")
